<?php

#------ Required Files -------#
require_once("class.ExtFields.php");
include_once("lib/transliterator/cTransliterator.php");

#------ OpenSRS Exception Class -------#
class OpenSRSException extends Exception {
    # Private variables
    private $response;

    /**
     * Construct Class Variables
     *
     * @param string $message: Error message
     * @param int $code: Error code
     * @param exception $previous: Exception message
     * @param array $options: Options
     **/
    public function __construct($message, $code = 0, Exception $previous = null, $options = array()) {
        parent::__construct($message, $code, $previous);
        $this->response = $options;
    }

    # GET, SET error messages
    public function getResponse() { return $this->response; }
    public function setResponse($response) { $this->response = $response; }
}

#------ OpenSRS Main Class -------#
class OpenSRS {

    # Define constants
    const TOP_REGISTER  = 1; ///< Registration Operation*/
    const TOP_TRANSFER  = 2; ///< Transfer Operation*/
    const TOP_RENEW     = 3; ///< Renew Operation*/
    const TOP_TERMINATE = 4; ///< Terminate Operation*/
    const TOP_RESOLVE   = 5; ///< Resolve Operation Manually*/

    const SUCCESS = 200;
    const ASYNC_REGISTRY = 250;

    # Private variables
    private $params;
    private $extData;
    private $domainName;
    private $contacts;
    private $autoRenewalTlds = array("de", "at","fr");

    /**
     * Construct Class Variables
     *
     * @param array $params: Data passed in from Odin
     **/
    public function __construct($params) {
        # Set submitted data
        $this->params = $params;

        # Set additional fields
        if (isset($params['additionalfields'])) {
            $this->extData = $params['additionalfields'];
        }

        # Set full root domain name
        if (isset($params['sld'])) {
            $this->domainName = $params['sld'] . '.' . $params['tld'];
        } elseif (isset($params['domainName'])) {
            $this->domainName = $params['domainName'];
        }

        # Set OpenSRS class lib data
        $this->initConst($params);

        # Set required files        
        require_once (__DIR__. "/src/opensrs/openSRS_loader.php");
    }

    /**
     * Define OpenSRS Class Library Connection Data
     *
     * @param array $params: Data passed in from Odin
     **/
    private function initConst($params) {
        # Set OpenSRS reseller username
        $osrs_username = $params["Username"];
        # Set OpenSRS LIVE server hostname
        $osrs_hostname = 'rr-n1-tor.opensrs.net';
        # Set OpenSRS LIVE API Key
        $osrs_key = $params["LIVE API Key"];

        # If test mode is enabled...
        if ($this->params['Test Mode'] === 'on'){
            # Set TEST system API key
            $osrs_key = $params["TEST API Key"];
            # SEt TEST server hostname
            $osrs_hostname = 'horizon.opensrs.net';
        }

        # If HRS is enabled...
        if ($this->params['HRS Customer'] == 'on'){
            # Change API server to HRS instance at given IP
            $osrs_hostname = $this->params['HRS API Server IP'];
        }

        # Define Opensrs class lib constants
        define('OSRS_USERNAME', $osrs_username);
        define('OSRS_KEY', $osrs_key);
        define('OSRS_HOST', $osrs_hostname);
    }

    /**
     * OpenSRS Odin Plugin Config
     *
     * @return array: Plugin config options
     **/
    public static function getConfig() {
        return array(
            "Username" => array("type" => "text", "size" => "20", "description" => "Enter your OpenSRS reseller ID. (Username)"),
            "LIVE API Key" => array("type" => "password", "size" => "20", "description" => "Enter your OpenSRS Live (Prod) API key here."),
            "TEST API Key" => array("type" => "password", "size" => "20", "description" => "Enter your OpenSRS Test (Horizon) API key here."),
            "Tech Contact" => array("type" => "yesno", "description" => "Use reseller's default Tech Contact info for domain registrations and transfers."),
            "ICANN TOS Link" => array("type" => "text", "size" => "50", "description" => "Link to your 
            Terms of Service (TOS) for Designated Agent (DA).", "placeholder" => "https://opensrs.com/wp-content/uploads/Tucows_ExhibitA.html"),

            "HRS Customer" => array("type" => "yesno", "description" => "This option changes server from OpenSRS to HRS API server."),
            "HRS API Server IP" => array("type" => "text", "size" => "20", "description" => "Enter HRS API Server IP."),

            "Test Mode" => array("type" => "yesno", "description" => "Run all API commands on OpenSRS test system. (Horizon)"),
            "Debug Mode" => array("type" => "yesno", "description" => "Sends output to: /var/log/httpd/pa-domain.log"),
            "Version" => array("type" => "dropdown","options" => "1.3|Version: 1.3"),
        );
    }

    /**
     * Call OpenSRS API Server
     *
     * @param string $function: OpenSRS lib function name
     * @param array $data: API function required data
     * @param array $custom: Custom contact data
     *
     * @return array $osrsResult: OpenSRS response in PHP array
     **/
    private function call($function, $data, $custom = array()) {
        # Construct request array
        $request = array(
            "func" => $function,
            "data" => $data,
        );

        # If custom data is provided, construct new merged array
        if (count($custom) > 0) $request = array_merge($request, $custom);

        # Debug mode: Request
        if ($this->params['Debug Mode'] === "on") {
            pa_log("RAW REQUEST\n".print_r($request, true));
        }

        # Process request
        $osrsHandler = processOpenSRS ("json", json_encode($request));

        # Debug mode: Response
        if ($this->params['Debug Mode'] === "on") {
            pa_log(print_r($osrsHandler, true));
        }

        # Format response to PHP array
        if (intval($osrsHandler->resultFullRaw["is_success"]) == 0) {
            $exception = new OpenSRSException($osrsHandler->resultFullRaw["response_text"]);
            $exception->setResponse($osrsHandler->resultFullFormatted);
            throw $exception;
        }

        # Covert JSON to PHP array
        $osrsResult = json_decode($osrsHandler->resultFormatted, true);

        # Return response
        return $osrsResult;
    }

    /**
     * Split Domain To TLD & SLD
     *
     * @param string $domain: Domain name
     *
     * @return array: Split domain name to TLD and SLD
     **/
    private function splitDomain($domain) {
        # Return domain split into TLD and SLD
        return explode('.', $domain, 2);
    }

    /**
     * Generate Unique OpenSRS Profile Username (20 Characters Max)
     *
     * @param string $domain: Domain name
     *
     * @return string: Unique opensrs profile username
     **/
    private function generateLogin($domain) {
        # Generate unique username for OpenSRS domain profile
        $username = mb_strimwidth(preg_replace('/[\.-]/', '', $domain), 0, 20);
        # Return unique username
        return $username;
    }

    /**
     * Generate Unique OpenSRS Profile Password (10 Characters Max)
     *
     * @param string $length: Maximum length of profile password
     *
     * @return string: Unique opensrs profile password
     **/
    private function generatePassword($length = 10) {
        # Set allowed characters
        $chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
        # Get character count
        $count = mb_strlen($chars);
        # Generate random password
        for ($i = 0, $result = ''; $i < $length; $i++) {
            $index = rand(0, $count - 1);
            $result .= mb_substr($chars, $index, 1);
        }
        # Return random password
        return $result;
    }

    /**
     * Checks Auto-Renewal Only TLDs
     *
     * @param string $tld: Auto-renewal check TLD
     *
     * @return bool: If TLD has auto-renew by default
     **/
    private function isAutoRenewalTld($tld) {
        # Return result of default auto-renewal status on TLD
        return in_array($tld, $this->autoRenewalTlds);
    }

    /**
     * Prepare Contact Sets
     **/
    public function prepareContacts() {
        # Construct contact sets for Registrant, Admin, Billing, Tech
        if (isset($this->params["contactdetails"])) {
            foreach(array("Registrant", "Billing", "Admin", "Tech") as $type) {
                # Set contact type
                $c = $this->params["contactdetails"][$type];

                # Convert contact type 'Registrant' to 'Owner'
                if ($type == "Registrant") {
                    $type = "Owner";
                }

                # Set contact set data
                $this->contacts[$type] = array(
                    "first_name" => $c["First Name"],
                    "last_name"  => $c["Last Name"],
                    "org_name" => $c["company"],
                    "phone" => $c["phonenumberseparated"],
                    "fax"	=> $c["faxseparated"],
                    "email" => $c["Email"],
                    "postal_code" => $c["postcode"],
                    "country"=> $c["country"],
                    "state" => $c["state"],
                    "city"	=> $c["city"],
                    "address1" => $c["address1"],
                    "address2" => $c["address2"],
                    "address3" => "",
                    "lang_pref"=> "EN",
                );

                # If company name is empty, set default to First and Last Name of Registrant
                if (empty($c["company"])) {
                    $this->contacts[$type]["org_name"] = $c["First Name"] ." ". $c["Last Name"];
                }

                # Clear memory
                unset($c);
            }
        } else {
            # If 'Registrant' contacts is missing...
            foreach	(array("", "billing", "tech", "admin") as $prefix) {
                # Set first character upper case
                $type = ucfirst($prefix);
                # Change blank entry to 'Owner'
                if ($type === "") {
                    $type = "Owner";
                }
                # Set contact data
                $c = $this->params;

                # Construct each contact set
                $this->contacts[$type] = array(
                    "first_name" => $c[$prefix."firstname"],
                    "last_name"  => $c[$prefix."lastname"],
                    "org_name" => $c[$prefix."company"],
                    "phone" => $c[$prefix."phonenumberseparated"],
                    "fax"	=> $c[$prefix."faxseparated"],
                    "email" => $c[$prefix."email"],
                    "postal_code" => $c[$prefix."postcode"],
                    "country"=> $c[$prefix."country"],
                    "state" => $c[$prefix."state"],
                    "city"	=> $c[$prefix."city"],
                    "address1" => $c[$prefix."address1"],
                    "address2" => $c[$prefix."address2"],
                    "address3" => "",
                    "lang_pref"=> "EN",
                );

                # If company name is missing, set default values to First and Last name
                if (empty($c[$prefix."company"])) {
                    $this->contacts[$type]["org_name"] = $c[$prefix."firstname"] ." ". $c[$prefix."lastname"];
                }

                # Clear memory
                unset($c);
            }
        }

        # Postprocessing
        foreach($this->contacts as $type => $c) {
            //E.164 Phone Number
            if (is_array($c["phone"])) {
                $phone = $c["phone"];
                $this->contacts[$type]["phone"] = "+".$phone["countryCode"].".".$phone["areaCode"].$phone["number"];
                if (!empty($phone["ext"])) $this->contacts[$type]["phone"] .= "x".$phone["ext"];
            } else {
                $this->contacts[$type]["phone"] = '';
            }
            //E.164 Fax Number
            if (is_array($c["fax"])) {
                $fax = $c["fax"];
                $this->contacts[$type]["fax"] = "+".$fax["countryCode"].".".$fax["areaCode"].$fax["number"];
                if (!empty($fax["ext"])) $this->contacts[$type]["fax"] .= "x".$fax["ext"];
            } else {
                $this->contacts[$type]["fax"] = '';
            }

            if (empty($this->contacts[$type]["org_name"]))
                $this->contacts[$type]["org_name"] = $this->contacts[$type]["first_name"] . " " . $this->contacts[$type]["last_name"];
        }
    }

    /**
     * Check Domain Availability
     *
     * @return array $lookup: Domain name and availability status
     */
    public function checkAvailability() {
        # Set variables
        $domainNames = $this->params["domainNames"];
        $slds = array();
        $tlds = array();

        # Loop through domain names submitted and split TLD and SLD
        foreach ($domainNames as $domain) {
            list($sld, $tld) = $this->splitDomain($domain);
            $tlds[] = ".$tld";
            $slds[] = $sld;
        }

        # Create a unique TLD and SLD arrays
        $slds = array_unique($slds);
        $tlds = array_unique($tlds);

        # Lookup availability status with OpenSRS for each domain
        $lookup = array();
        foreach($slds as $sld) {
            $response = $this->call("lookupDomain",
                array(
                    "domain" => $sld,
                    "selected" =>  implode(';', $tlds),
                )
            );

            # Debug mode
            if ($this->params['Debug Mode'] === "on") {
                pa_log("DOMAIN LOOKUP::RESPONSE\n".print_r($domainNames,true));
            }

            # Construct availability status array
            foreach($response as $i => $item) {
                $status = ($item["status"] == "available") ? 1 : 2;
                $msg = "Domain is ".$item["status"];
                $lookup[$item["domain"]] = array("status" => $status, "msg" => $msg);
            }

        }

        # Return lookup results
        return $lookup;
    }

    /**
     * Gets Domain Order Info
     *
     * @param string $orderId: OpenSRS domain order ID
     *
     * @return array $response: Order info
     */
    public function getOrderInfo($orderId) {
        # Get order info from given ID
        $response = $this->call("lookupGetOrderInfo",
            array(
                "order_id" => $orderId,
            )
        );
        # Return order info
        return $response;
    }

    /**
     * Modify Auto Renew Setting
     *
     * @param string $autoRenew: True or false value for auto-renew
     *
     * @return array $response: Return response
     */
    public function modifyAutoRenew($autoRenew) {
        # Modify auto-renew setting
        try {
            $response = $this->call("provModify",
                array(
                    "domain" => $this->domainName,
                    "data" => "expire_action",
                    "auto_renew" => ($autoRenew) ? "1": "0",
                    "let_expire" => "0",
                )
            );
            # Return success message
            return array("success" => true, "msg" => "Auto renew setting modified.");
        } catch (OpenSRSException $e) {
            # Catch exceptions and output error
            return array('error' => 'Unable to modify auto_renew (Error: "'.$e->getMessage().'. ');
        }
    }

    /**
     * Get Expiration Date
     *
     * @return array $response: Return status and expiration date
     */
    public function getExpirationDate() {
        # Get expiration date on given domain
        try {
            $response = $this->call("lookupBelongsToRsp",
                array(
                    "domain" => $this->domainName,
                )
            );

            # If domain does not belong to reseller, output error
            if (isset($response["belongs_to_rsp"]) && $response["belongs_to_rsp"] == "1") {
                # Return expiration date if it's set...
                if (!empty($response["domain_expdate"])) {
                    return array("success" => true, "expirydate" => $response["domain_expdate"] );
                }
                # Else return empty expiration date
                return array("success" => true, "expirydate" => "");

            }
            # Output error if domain does not belong to reseller
            throw new OpenSRSException("Domain does not belong to the RSP");

        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to get expiration date (Error: "'.$e->getMessage().'.');
        }
    }

    /**
     * Sync Domain Order Status
     *
     * @return array: Return domain order status
     */
    public function syncDomainRegistration() {
        # Set domain order ID
        $orderId = $this->extData["OrderID"];

        # Get domain order info
        try {
            # Set order ID
            $order = $this->getOrderInfo($orderId);

            # Debug mode: Request order details
            if ($this->params['Debug Mode'] === "on") {
                pa_log("SYNC RESPONSE\n".print_r($order, true));
            }

            # Return response based on declined order status
            if ($order['attributes']['field_hash']['status']=='canceled'
                || $order['attributes']['field_hash']['status']=='declined'
                || $order['attributes']['field_hash']['status']=='deleted') {
                return array("error" => $order["response_text"]);
            }

            # Return response if successful
            if ($order['attributes']['field_hash']['status']=='completed' ) {
                # Get expiry date
                $expirydate = $this->getExpirationDate();

                return array("completed" => true, 'expirydate' => $expirydate['expirydate']);
            } else {
                # Get expiry date
                $expirydate = $this->getExpirationDate();

                # Else return failed status
                return array("completed" => false, 'expirydate' => $expirydate['expirydate']);
            }
        } catch (OpenSRSException $e) {
            # Catch exceptions and return
            return array('error' => "There was a connection error to registrar. Please contact support for assistance.
        	Error: ".$e->getMessage());
        }
    }

    /**
     * Validate Additional Data Feilds
     *
     * @return array $tldFeilds: Return additional TLD data
     */
    public function validateAdditionalFields() {
        # Create new instance
        $extFields = new ExtFields(
            $this->params["sld"],
            $this->params["tld"],
            $this->params["operationType"],
            $this->params["additionalfields"],
            null,
            $this->params['ICANN TOS Link']
        );
        # Set additional required fields for given TLD
        $tldFields = $extFields->getFields();
        # Return additional required fields
        return $tldFields;
    }

    /**
     * Check Transfer Availability
     *
     * @return array: Return transfer availability status of domain
     */
    public function checkTransferAvailability() {
        # Check if domain is available for transfer
        try {
            # Call OpenSRS API
            $response = $this->call("transCheck",
                array(
                    "domain" => $this->domainName,
                )
            );

            # Return transfer status
            if (isset($response['transferrable']) && intval($response['transferrable']) != 0) {
                return array('status' => 1, 'renewRequired' => 0, 'msg' => 'This domain is available for transfer');
            } else {
                return array('status' => 2, 'renewRequired' => 0, 'msg' => 'This domain is not available for transfer');
            }
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => $e->getMessage());
        }
    }

    /**
     * Set Additional Required Fields
     *
     * @return array $tldFeilds: Return additional TLD data
     */
    public function fillExtTldData() {
        # Initiate variables
        $tld = $this->params["tld"];
        $data = array();

        #---- Set specific TLD data fields ----#
        #.PRO
        if ($tld == 'pro') {
            $map = array(
                "PRO_Authority" => "authority",
                "PRO_AuthorityWebsite" => "authority_website",
                "PRO_LicenseNumber" => "license_number",
                "PRO_Profession" => "profession",
            );

            foreach ($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("professional_data" => $data);
        }

        #.ASIA
        if ($tld == 'asia') {
            $map = array(
                "ASIA_ContactType" => "contact_type",
                "ASIA_IdNumber" => "id_number",
                "ASIA_IdType" => "id_type",
                "ASIA_LegalEntityType" => "legal_entity_type",
                "ASIA_LocalityCountry" => "locality_country",
            );

            if($this->extData['ASIA_IdType'] == "other"){
                $idTypeDes = array("ASIA_IdTypeInfo" => "id_type_info");
                array_merge($map, $idTypeDes);
            }

            if($this->extData['ASIA_LegalEntityType'] == "other"){
                $legalTypeDes = array("ASIA_LegalEntityTypeInfo" => "legal_entity_type_info");
                array_merge($map, $legalTypeDes);
            }

            foreach ($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("cedinfo" => $data);
        }

        #.COM.AU, .NET.AU, .ORG.AU, .ASN.AU, .ID.AU
        if (in_array($tld, array("com.au", "net.au", "org.au", "asn.au", "id.au"))) {
            $map = array(
                "AU_RegistrantName" => "registrant_name",
                "AU_RegistrantIdType" => "registrant_id_type",
                "AU_RegistrantId" => "registrant_id",
                "AU_EligibilityType" => "eligibility_type",
                "AU_EligibilityName" => "eligibility_name",
                "AU_EligibilityIdType" => "eligibility_id_type",
                "AU_EligibilityId" => "eligibility_id"
            );

            foreach($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("au_registrant_info" => $data);
        }

        #.BE
        if ($tld == 'be') {
            $map = array(
                "OwnerConfirmAddress" => "owner_confirm_address",
                "BE_Lang" => "lang"
            );

            foreach ($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("registrant_extra_info" => $data);
        }

        #.US
        if ($tld == 'us') {
            $map = array(
                "US_AppPurpose" => "app_purpose",
                "US_Category" => "category",
                "US_Validator" => "validator"
            );

            foreach($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("nexus" => $data);
        }

        #.CA
        if ($tld == 'ca') {
            $map = array(
                "CA_LegalType" => "legal_type",
                "CA_LangPref" => "lang_pref"
            );

            foreach ($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("registrant_extra_info" => $data);
        }

        #.IT
        if ($tld == 'it') {
            $map = array(
                "IT_EntityType" => "entity_type",
                "IT_NationalityCode" => "nationality_code",
                "IT_RegistrantCode" => "reg_code"
            );

            foreach($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("it_registrant_info" => $data);
        }

        #.FR
        if ($tld == 'fr'){
            $map = array(
                "FR_RegistrantType" => "registrant_type ",
                "FR_CountryOfBirth" => "country_of_birth",
                "FR_DateOfBirth" => "date_of_birth",
                "FR_PlaceOfBirth" => "place_of_birth",
                "FR_PostalCodeOfBirth" => "postal_code_of_birth",
                "FR_BusinessRegNumber" => "siren_siret",
                "FR_TrademarkNumber" => "trademark_number",
                "FR_RegistrantVatId" => "registrant_vat_id"
            );

            foreach($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("registrant_extra_info" => $data);
        }

        #.SE / .NU
        if (in_array($tld, array('se', 'nu'))){
            $map = array(
                "SE_IdCardNumber" => "id_card_number",
                "SE_RegistrantType" => "registrant_type",
                "SE_RegistrantVatId" => "registrant_vat_id",
                "SE_RegistrationNumber" => "registration_number",
            );

            foreach($map as $oa => $osrs) {
                if (!empty($this->extData[$oa])) $data[$osrs] = $this->extData[$oa];
            }

            return array("registrant_extra_info" => $data);
        }
    }

    /**
     * Register Domain Name
     *
     * @return array: Return registration order response
     */
    public function registerDomain() {
        # Initiate variables
        $regPeriod = $this->params["regperiod"];
        $nameservers = array();
        $orderID = 0;

        # Set name servers
        for ($i = 1; $i <= 5; $i++) {
            if (isset($this->params['ns' . $i])) {
                $nameservers['name' . $i] = $this->params['ns' . $i];
                $nameservers['sortorder' . $i] = $i;
            }
        }

        # Construct OpenSRS profile username/password
        $login = $this->generateLogin($this->domainName);
        $password = $this->generatePassword();

        # Get additional data if required
        $tldData = $this->fillExtTldData();

        # Construct contact sets
        $this->prepareContacts();
        $owner = $this->contacts["Owner"];
        $admin = $this->contacts["Admin"];
        $billing = $this->contacts["Billing"];
        $tech = $this->contacts["Tech"];

        # Check for custom Tech contact
        $useCustomerTechContact = ($this->params["Tech Contact"] == "on") ? 1 : 0;

        # Set international lang if required
        if (isset($this->extData["IdnLang"])) {
            $idnLang = $this->extData["IdnLang"];
        }

        # Get TLD
        $tld = $this->params["tld"];

        # Get user IP address
        $user_ip = $_SERVER['REMOTE_ADDR'];

        # Try domain registration
        try {
            # Set common data set for all TLDs
            $request = array(
                "domain" => $this->domainName,
                "auto_renew" => $this->isAutoRenewalTld($tld),
                "f_lock_domain" => "0",
                "custom_nameservers" => "1",
                "prereg" => "0",
                "period" => $regPeriod,
                "reg_type" => "new",
                "custom_tech_contact" => ($useCustomerTechContact) ? "1" : "0",
                "reg_username" => $login,
                "reg_password" => $password,
            );

            # Set international language if enabled
            if (!empty($idnLang)) $request["encoding_type"] = $idnLang;

            # Set Contact sets
            $custom = array(
                "personal" => $owner,
                "admin" => $admin,
                "tech" => $tech,
                "billing" => $billing
            );

            # Merge request
            $request = array_merge((array)$request, (array)$nameservers);

            # Construct new contact sets and additional required data array
            $custom  = array_merge((array)$custom, (array)$tldData);

            # Process registration
            $response = $this->call("provSWregister", $request, $custom);

            # Debug mode: Registration response
            if ($this->params['Debug Mode'] === "on") {
                pa_log("REGISTRATION RESPONSE\n".print_r($response,true));
            }

            # If successful, set Order ID
            if ($response["response_code"] == self::SUCCESS || $response["response_code"] == self::ASYNC_REGISTRY) {
                # Set order ID
                $orderID = $response["attributes"]["id"];
                # Set designated agent (DA) state
                $da_data = array(
                    'domain' => $this->domainName,
                    'enable' => True,
                    'ip_address' => $user_ip
                );
                $da_accepted = $this->call('modifyTradeDAStatus', $da_data);

                # Log registrant acceptance for Tucows Inc. to act as their designated agent (DA) for ICANN audits
                pa_log("DA ACCEPTED::REGISTRATION::REQUEST::\n".print_r($da_data,true));
                pa_log("DA ACCEPTED::REGISTRATION::RESPONSE::\n".print_r($da_accepted,true));

            }
        } catch (OpenSRSException $e) {
            # Catch errors and set pending Order ID
            if (isset($response["forced_pending"]) && !empty($response["force_pending"])) {
                $orderID = $response["force_pending"];
            }
        }

        # Return success data
        if ($orderID) {
            return array(
                "additionalfields" => array(
                    "OrderID" => $orderID,
                    "Login" => $login,
                    "Passwd" => $password,
                ),
            );
        } else {
            # Return error message
            return array('error' => "Sorry! Domain registration failed at registrar. Please contact support for 
    		assistance.");
        }
    }

    /**
     * Transfer Domain Name
     *
     * @return array: Return transfer availability status of domain
     */
    public function transferDomain() {
        # Initiate variables
        $tld = $this->params["tld"];
        $regPeriod = $this->params["regperiod"];
        $nameservers = array();
        $orderID = 0;

        # Handle special transfer period TLDs
        $specialPeriodTLDs = ['au'];
        $skipTransferPeriod = false;
        if (in_array($tld, $specialPeriodTLDs)){
            $skipTransferPeriod = true;
        }

        # Set name servers
        for ($i = 1; $i <= 5; $i++) {
            if (isset($this->params['ns' . $i])) {
                $nameservers['name' . $i] = $this->params['ns' . $i];
                $nameservers['sortorder' . $i] = $i;
            }
        }

        # Set OpenSRS profile username
        $login = $this->generateLogin($this->domainName);
        # Set OpenSRS profile password
        $password = $this->generatePassword();
        # Set additional fields
        $tldData = $this->fillExtTldData();

        # Prepare contact sets
        $this->prepareContacts();
        $owner = $this->contacts["Owner"];
        $admin = $this->contacts["Admin"];
        $billing = $this->contacts["Billing"];
        $tech = $this->contacts["Tech"];

        # Set reseller tech contact address as default
        $useCustomerTechContact = ($this->params["TechContact"] == "on") ? 1 : 0;;

        # Set international language
        if (isset($this->extData["IdnLang"])) {
            $idnLang = $this->extData["IdnLang"];
        }

        # Transfer domain name
        try {
            $request = array(
                "domain" => $this->domainName,
                "auto_renew" => $this->isAutoRenewalTld($this->params["tld"]),
                "f_lock_domain" => "0",
                "custom_nameservers" => "1",
                "custom_transfer_nameservers" => "1",
                "prereg" => "0",
                "period" => ($skipTransferPeriod) ? "0" : $regPeriod,
                "reg_type" => "transfer",
                "custom_tech_contact" => ($useCustomerTechContact) ? "1" : "0",
                "reg_username" => $login,
                "reg_password" => $password,
            );

            # Set international language if enabled
            if (!empty($idnLang)) $request["encoding_type"] = $idnLang;

            # Set Contact sets
            $custom = array(
                "personal" => $owner,
                "admin" => $admin,
                "tech" => $tech,
                "billing" => $billing
            );

            # Set custom tech contact if enabled
            if ($useCustomerTechContact) $custom["tech"] = $tech;

            # Set custom contact type for .DE
            if ($tld == "de") $custom["zone"] = $tech;

            # Merge request
            $request = array_merge((array)$request, (array)$nameservers);

            # Construct new contact sets and additional required data array
            $custom  = array_merge((array)$custom, (array)$tldData);

            # Process transfer order
            $response = $this->call("provSWregister", $request, $custom);

            # Debug mode: Registration response
            if ($this->params['Debug Mode'] === "on") {
                pa_log("TRANSFER RESPONSE\n".print_r($response,true));
            }

            # If successful, set Order ID
            if ($response['attributes']["registration_code"] ==
                self::SUCCESS || $response["registration_code"] == self::ASYNC_REGISTRY) {
                $orderID = $response['attributes']["id"];
            }
        } catch (OpenSRSException $e) {
            # Catch errors and set pending Order ID
            if (isset($response["forced_pending"]) && !empty($response["force_pending"])) {
                $orderID = $response["force_pending"];
            }
        }

        # Return data on success
        if ($orderID) {
            return array(
                "additionalfields" => array(
                    "OrderID" => $orderID,
                    "Login" => $login,
                    "Passwd" => $password,
                ),
            );
        } else {
            #Return data on failure
            return array('error' => "Domain transfer in failed.");
        }
    }

    /**
     * Sync Domain Transfer Order
     *
     * @return array: Return transfer order details
     */
    public function syncDomainTransfer() {
        return $this->syncDomainRegistration();
    }

    /**
     * Modify WHOIS privacy
     *
     * @return array: Return modification status
     */
    public function setWhoisPrivacy() {
        # Set WHOIS privacy setting
        $protectenable = $this->params["protectenable"];

        # Debug mode: Output WHOIS privacy setting
        if ($this->params['Debug Mode'] === "on") {
            pa_log("ID PROTECT FLAG\n".print_r($protectenable, true));
        }

        # Modify WHOIS privacy
        try {
            $response = $this->call("provModify",
                array(
                    "domain" => $this->domainName,
                    "affect_domains" => "0",
                    "report_email" => "",
                    "data" => "whois_privacy_state",
                    "state" => $protectenable,
                )
            );
            #Return status
            return array("success" => true);
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to modify WhoIs privacy: ("'.$e->getMessage().'. ');
        }
    }

    /**
     * Modify Registrar Lock
     *
     * @return array: Return modification status
     */
    public function saveRegistrarLock() {
        # Set default locked state
        $lockState = "0";
        # Set locked state
        if (isset($this->params['lockenabled']) && $this->params['lockenabled'] == 'locked') {
            $lockState = "1";
        }

        # Modify registrar lock
        try {
            $response = $this->call("provModify",
                array(
                    "domain" =>  $this->domainName,
                    "affect_domains" => "0",
                    "data" => "status",
                    "lock_state" => $lockState,
                    "domain_name" => $this->domainName,
                )
            );
            # Return status
            return array("success" => true);
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to modify Registrar Lock state: ("'.$e->getMessage().'. ');
        }
    }

    /**
     * Register Custom Name Servers
     *
     * @param string $nameserver: Name server to register
     *
     * @return array: Return name server registration status
     */
    public function registerNameServers($nameserver) {
        # Register custom name servers with all registries
        try {
            $response = $this->call("nsRegistryAdd",
                array(
                    "all" => "1",
                    "fqdn" => $nameserver,
                    "tld" => $this->params["tld"],
                )
            );
            # Return status
            return array("success" => true);
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to register nameserver (Error: "'.$e->getMessage().'). ');
        }
    }

    /**
     * Get Name Server Status
     *
     * @return array: Return status
     */
    public function getNameserversStatus() {
        return array("completed" => true);
    }

    /**
     * Check Name Servers
     *
     * @return array: Output name server check status
     */
    public function checkNameservers() {

        # Construct name server array
        $nameservers = array();
        for ($i = 1; $i <= 5; $i++) {
            if (isset($this->params['ns' . $i])) {
                $nameservers[] = $this->params['ns' . $i];
            }
        }

        # Get domain's TLD
        $domain = explode ('.', $this->domainName);
        $tld = $domain[1];

        #Check each name server with corrasponding registry
        foreach ($nameservers as $nameserver) {
            try {
                # Check name server
                $response = $this->call("nsRegistryCheck",
                    array(
                        "fqdn" => $nameserver,
                        "tld" => ".".$tld,
                    )
                );

                # Debug mode: Check name servers output
                if ($this->params['Debug Mode'] === "on") {
                    pa_log("REQUEST DOMAIN NAME\n".print_r($domain, true));
                    pa_log("REQUEST CHECK NAMESERVER\n".print_r($nameservers, true));
                    pa_log("RESPONSE CHECK NAMESERVER\n".print_r($response, true));
                }
                # Return status
                return array("success" => true);
            } catch (OpenSRSException $e) {
                # Debug mode: Output request info on error
                if ($this->params['Debug Mode'] === "on") {
                    pa_log("REQUEST DOMAIN NAME\n".print_r($domain, true));
                    pa_log("REQUEST NAMESERVERS\n".print_r($nameservers, true));
                }

                # Register name server that don't exist at registry database
                $regNSStatus = $this->registerNameServers($nameserver);

                # Output name server check status
                if ($regNSStatus["success"]){
                    return array("success" => true);
                } else {
                    return array('error' => 'Unable to modify name servers: ("'.$e->getMessage().'")');
                }
            }
        }
    }

    /**
     * Save Name Servers
     *
     * @return array: Output name server check status
     */
    public function saveNameservers() {

        //XPACK tlds: CH, FR, NL, AT, COM.MX, ES, DK, LI.
        //For XPACk tlds method update_all_info should be used instead of advanced_update_nameservers.

        # Construct name servers
        $nameservers = array();
        for ($i = 1; $i <= 5; $i++) {
            if (isset($this->params['ns' . $i])) {
                $nameservers[] = $this->params['ns' . $i];
            }
        }

        # Assign name servers to domain	
        try {
            $response = $this->call("nsAdvancedUpdt",
                array(
                    "domain" => $this->domainName,
                    "op_type" => "assign",
                    "assign_ns" => implode(',', $nameservers),
                )
            );

            # Debug mode: Output assign name servers response
            if ($this->params['Debug Mode'] === "on") {
                pa_log("ASSIGN NAMESERVERS RESPONSE\n".print_r($response, true));
            }

            # Return status
            return array("success" => true);
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to modify name servers: ("'.$e->getMessage().'").');
        }
    }

    /**
     * Sync Domain Renewal Date
     *
     * @return array: Domain info
     */
    public function syncDomainRenewal() {
        return $this->syncDomainRegistration();
    }

    /**
     * Renew Domain
     *
     * @return array: Return domain renewal status
     */
    public function renewDomain() {
        # Check if TLD is by default set to auto-renew
        $autoRenew = $this->isAutoRenewalTld($this->params["tld"]);
        # Set renew period
        $renewalPeriod = $this->params["regperiod"];

        # If TLD is by default auto-renew, modify auto-renew setting
        if ( $autoRenew ) {
            # Return modify auto-renew setting response
            return  $this->modifyAutoRenew($autoRenew);
        }

        # Get current expiration date
        $response = $this->getExpirationDate();

        # Debug mode: Get current expiration date response
        if ($this->params['Debug Mode'] === "on") {
            pa_log("GET CURRENT EXPIRATION DATE RESPONSE\n".print_r($response, true));
        }

        # Set current expiration date based on response, else return error
        if (isset($response["error"])) {
            return $response;
        } elseif (isset($response["expirydate"]) && sscanf($response["expirydate"], "%04u-%02u-%02u", $year, $month, $day) > 0) {
            $currentExpirationYear = $year;
        } else {
            return array("error" => "Failed to get current expiration date");
        }

        # Initiate variables
        $orderID = 0;
        $requestAccepted = false;
        $errorMessage = false;

        # Renew domain name
        try {
            $response = $this->call("provRenew",
                array(
                    "domain" => $this->domainName,
                    "currentexpirationyear" => "$currentExpirationYear",
                    "period" => $renewalPeriod,
                    "auto_renew" =>  "0",
                    "handle" => "process",
                )
            );

            # Set active order ID
            if (isset($response["order_id"])) {
                $orderID = $response["order_id"];
                $requestAccepted = true;
            }
        } catch (OpenSRSException $e) {
            # Catch errors and set pending order ID
            if (isset($response["forced_pending"])) {
                $orderID = $response["forced_pending"];
                $requestAccepted = true;
            }
            # Set error message
            $errorMessage = $e->getMessage();
        }

        # Return status
        if (!$requestAccepted) {
            return array("error" => "Domain renewal in OpenSRS failed: $errorMessage");
        } else {
            return array("additionalfields" => array("OrderID" => $orderID),);
        }
    }

    /**
     * Update Domain Contact Set
     * @return array: Return contact update status
     */
    public function updateDomainContacts() {
        # Prepare contact sets
        $this->prepareContacts();
        $owner = $this->contacts["Owner"];
        $admin = $this->contacts["Admin"];
        $billing = $this->contacts["Billing"];
        $tech = $this->contacts["Tech"];

        # Format contact type name for specific TLDs
        $types = array("owner", "tech", "admin");
        if ($this->params["tld"] != "ca" && $this->params["tld"] != "nl") $types[] = "billing";

        # Get Current Owner contact set from OpenSRS
        $owner_contact = $this->call('lookupGetDomain',array(
            'domain' => $this->domainName,
            'type' => 'owner'
        ));

        # Get user IP address
        $user_ip = $_SERVER['REMOTE_ADDR'];

        # Set ICANN trade conditions
        $trade_conditions = array('first_name', 'last_name', 'org_name', 'email');

        # Initialize variables
        $trade_required = False;

        # $owner_contact['attributes']['contact_set']['owner']['']
        foreach ($owner_contact['attributes']['contact_set']['owner'] as $field => $value) {
            # Check if field is valid ICANN trade field
            if(in_array($field, $trade_conditions)) {
                # Validate value
                if($owner[$field] != $value) {
                    $trade_required = True;
                    break;
                }
            }
        }

        # GC::Check trade required
        if($trade_required == True) {
            # Get designated Agent (DA) status
            $da_enabled = $this->call('getTradeDAStatus', array('domain' => $this->domainName));

            # GC::Check DA status
            if ($da_enabled['attributes']['allow_designated_agent'] == True) {
                # Set new contact info
                $new_contact = array(
                    'enable' => True,
                    'ip_address' => $user_ip,
                    'first_name' => $owner['first_name'],
                    'last_name' => $owner['last_name'],
                    'org_name' => $owner['org_name'],
                    'email' => $owner['email']
                );
                # Enable new registrant DA state
                $new_contact_da_status = $this->call('enableDATrade', $new_contact);

                # Log new registrant DA enabled status for ICANN audits
                pa_log("NEW REGISTRANT DA ENABLE::REQUEST\n" . print_r($new_contact, true));
                pa_log("NEW REGISTRANT DA ENABLE::RESPONSE\n" . print_r($new_contact_da_status, true));

                # GC::Check Response
                if($new_contact_da_status['is_success'] == True) {
                    # Process trade
                    $contact_update = $this->call("provUpdateContacts",
                        array(
                            "domain" => $this->domainName,
                            "types" => implode(",", $types),
                        ),
                        array(
                            "personal" => $owner,
                            "admin" => $admin,
                            "billing" => $billing,
                            "tech" => $tech,
                        )
                    );

                    # Debug mode
                    if ($this->params['Debug Mode'] === "on") {
                        pa_log("DA ACTIVE::UPDATE::\n" . print_r($contact_update, true));
                    }

                    # GC::Handle Response
                    if($contact_update['is_success'] == True) {
                        # Return success
                        return array('success' => True);
                    } else {
                        # Return error
                        return array("error" => "Contact update failed at registrar. Please contact support 
                        for assistance.");
                    }
                } else {
                    # Update contact, however approval requires from new registrant
                    $contact_update = $this->call("provUpdateContacts",
                        array(
                            "domain" => $this->domainName,
                            "types" => implode(",", $types),
                        ),
                        array(
                            "personal" => $owner,
                            "admin" => $admin,
                            "billing" => $billing,
                            "tech" => $tech,
                        )
                    );

                    # Debug mode
                    if ($this->params['Debug Mode'] === "on") {
                        pa_log("APPROVAL REQUIRED::TRADE INITIATED::\n" . print_r($contact_update, true));
                    }

                    # GC::Handle Response
                    if($contact_update['is_success'] == True) {
                        # Return success
                        return array('success' => True);
                    } else {
                        # Return error
                        return array("error" => "Contact update failed at registrar. Please contact support 
                        for assistance.");
                    }
                }
            } else {
                # Enable DA status for 'old' registrant
                $request = array('domain' => $this->domainName, 'enable' => True, 'ip_address' => $user_ip);
                $da_modified = $this->call('modifyTradeDAStatus', $request);

                # Log old registrant DA enabled status for ICANN audits
                pa_log("DA ENABLE::REQUEST\n" . print_r($request, true));
                pa_log("DA ENABLE::RESPONSE\n" . print_r($da_modified, true));

                # GC::DA Status enabled
                if ($da_modified['is_success'] == True) {
                    # Set new contact info
                    $new_contact = array(
                        'enable' => True,
                        'ip_address' => $user_ip,
                        'first_name' => $owner['first_name'],
                        'last_name' => $owner['last_name'],
                        'org_name' => $owner['org_name'],
                        'email' => $owner['email']
                    );
                    # Enable new registrant DA state
                    $new_contact_da_status = $this->call('enableDATrade', $new_contact);

                    # Log new registrant DA enabled status for ICANN audits
                    pa_log("NEW REGISTRANT DA ENABLE::REQUEST\n" . print_r($request, true));
                    pa_log("NEW REGISTRANT DA ENABLE::RESPONSE\n" . print_r($da_modified, true));

                    # GC::Check Response
                    if($new_contact_da_status['is_success'] == True) {
                        # Process trade
                        $contact_update = $this->call("provUpdateContacts",
                            array(
                                "domain" => $this->domainName,
                                "types" => implode(",", $types),
                            ),
                            array(
                                "personal" => $owner,
                                "admin" => $admin,
                                "billing" => $billing,
                                "tech" => $tech,
                            )
                        );

                        # Debug mode
                        if ($this->params['Debug Mode'] === "on") {
                            pa_log("DA MODIFIED::UPDATE::\n" . print_r($contact_update, true));
                        }

                        # GC::Check trade status
                        if($contact_update == True) {
                            $transfer_lock = $this->call('modifyTradeLock', $request);

                            # Debug mode
                            if ($this->params['Debug Mode'] === "on") {
                                pa_log("TRANSFER LOCK DISABLE::RESPONSE\n" . print_r($transfer_lock, true));
                            }

                            # Return status
                            if($transfer_lock['is_success'] == True) {
                                # Return success
                                return array("success" => True);
                            } else {
                                # Return Success
                                return array('error' => "Contact info successfully updated, however unable to 
                            opt-out of 60 day transfer lock. Domain has been locked for 60 days from being transferred.");
                            }
                        } else {
                            # Return error message
                            return array("error" => "Contact update failed at registrar. Please contact support 
                        for assistance.");
                        }
                    } else {
                        # Update contact, however approval requires from new registrant
                        $contact_update = $this->call("provUpdateContacts",
                            array(
                                "domain" => $this->domainName,
                                "types" => implode(",", $types),
                            ),
                            array(
                                "personal" => $owner,
                                "admin" => $admin,
                                "billing" => $billing,
                                "tech" => $tech,
                            )
                        );

                        # Debug mode
                        if ($this->params['Debug Mode'] === "on") {
                            pa_log("APPROVAL REQUIRED::TRADE INITIATED::\n" . print_r($contact_update, true));
                        }

                        # GC::Handle Response
                        if($contact_update['is_success'] == True) {
                            # Return success
                            return array('success' => True);
                        } else {
                            # Return error
                            return array("error" => "Contact update failed at registrar. Please contact support 
                        for assistance.");
                        }
                    }
                } else {
                    # Initiate trade
                    $contact_update = $this->call("provUpdateContacts",
                        array(
                            "domain" => $this->domainName,
                            "types" => implode(",", $types),
                        ),
                        array(
                            "personal" => $owner,
                            "admin" => $admin,
                            "billing" => $billing,
                            "tech" => $tech,
                        )
                    );

                    # Debug mode
                    if ($this->params['Debug Mode'] === "on") {
                        pa_log("APPROVAL REQUIRED::TRADE INITIATED::\n" . print_r($contact_update, true));
                    }

                    # GC::Handle Response
                    if($contact_update['is_success'] == True) {
                        # Return Success::Require Approval Message
                        return array("success" => "Trade approval required. Approval emails have been sent to both old 
                        and new registrant.");
                    } else {
                        # Return error
                        return array("error" => "Contact update failed at registrar. Please contact support 
                        for assistance.");
                    }
                }
            }
        } else {
            # Update contact sets
            $contact_update = $this->call("provUpdateContacts",
                array(
                    "domain" => $this->domainName,
                    "types" => implode(",", $types),
                ),
                array(
                    "personal" => $owner,
                    "admin" => $admin,
                    "billing" => $billing,
                    "tech" => $tech,
                )
            );

            # Debug mode
            if ($this->params['Debug Mode'] === "on") {
                pa_log("STANDARD::UPDATE::\n" . print_r($contact_update, true));
            }

            # Return status
            if ($contact_update['is_success'] == True) {
                return array("success" => True);
            } else {
                # Return error message
                return array("error" => "Contact update failed at registrar. Please contact support for assistance.");
            }
        }
    }

    /**
     * Get Auth Key (EPP Code)
     *
     * @return array: Return auth code
     */
    public function getAuthKey() {
        # Get domain's auth code
        try {
            $response = $this->call("lookupGetDomain",
                array(
                    "domain" => $this->domainName,
                    "type" => "domain_auth_info",
                )
            );

            # Debug mode: Get auth code repsonse
            if ($this->params['Debug Mode'] === "on") {
                pa_log("GET EPP REPONSE\n".print_r($response, true));
            }

            # Return auth code or error based on response status
            if (isset($response['attributes']["domain_auth_info"])) {
                return array("eppcode" => $response['attributes']["domain_auth_info"]);
            } else {
                return array('error' => 'Unable to get auth info for domain.');
            }
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array('error' => 'Unable to get auth info for domain (Error: "'.$e->getMessage().'"). ');
        }
    }

    /**
     * Revoke Domain
     *
     * @return array: Return revoke status
     */
    public function revokeDomain() {
        # Revoke domain name
        try {
            $response = $this->call("provRevoke",
                array(
                    "reseller" => $this->params['Username'],
                    "domain" => $this->domainName,
                    "notes" => "Domain Revoked",
                )
            );

            # Debug mode: Revoke domain reponse
            if ($this->params['Debug Mode'] === "on") {
                pa_log("REVOKE RESPONSE\n".print_r($response, true));
            }

            # Return status
            return array(
                "success" => true,
                "additionalfields" => array("RevokeStatus" => "done")
            );
        } catch (OpenSRSException $e) {
            # Catch errors and return
            return array(
                "error" => $e->getMessage(),
                "additionalfields" => array("RevokeError" => $e->getMessage())
            );
        }
    }

    /**
     * Sync Revoke Order
     *
     * @return array: Return revoke status
     */
    public function syncDomainRevocation() {
        # TODO: Build revoke function for admin
        if (isset($this->extData["RevokeStatus"]) && $this->extData["RevokeStatus"] == "done") {
            return array("completed" => true);
        } else {
            return array("error" => $this->extData["RevokeError"]);
        }
    }

}